"""
bluetooth module
"""

# file: bluetooth.py
# auth: Albert Huang <albert@csail.mit.edu>
# desc: wrappers around the _bluetooth native python extension
# $Id: bluetooth.py,v 1.3 2005/04/01 01:08:17 albert Exp $

import _bluetooth
import _bluetooth as bt
import struct

btsock = bt.btsocket
HCI    = bt.HCI
RFCOMM = bt.RFCOMM
L2CAP  = bt.L2CAP
SCO    = bt.SCO

class BluetoothError(Exception):
    pass

def discover_devices(duration=8, flush_cache=True):
    """
    performs a bluetooth device discovery using the first available bluetooth
    resource.

    returns a list of bluetooth addresses.

    duration=8 
        how long, in units of 1.28 seconds, to search for devices.  To find as
        many devices as possible, you should set this to at least 8.
    flush_cache=True 
        if set to False, then discover_devices may return devices found during
        previous discoveries.
    """

    sock = _gethcisock()
    try:
        results = bt.hci_inquiry(sock, duration=8, flush_cache=True)
    except bt.error:
        sock.close()
        raise BluetoothError("error communicating with local bluetooth device")

    sock.close()
    return results

def lookup_name(address, timeout=10):
    """
    Tries to determine the friendly name (human readable) of the device with
    the specified bluetooth address.  Returns the name on success, and None
    on failure.

    timeout=10   how many seconds to search before giving up.
    """
    _checkaddr(address)

    sock = _gethcisock()
    timeoutms = int(timeout * 1000)
    try: 
        name = bt.hci_read_remote_name( sock, address, timeoutms )
    except bt.error, e:
        # name lookup failed.  either a timeout, or I/O error
        name = None
    sock.close()
    return name

def is_valid_address(address):
    """
    returns True if address is a valid bluetooth address

    valid address are always strings of the form XX:XX:XX:XX:XX:XX
    where X is a hexadecimal character.  For example,
        01:23:45:67:89:AB is a valid address, but
        IN:VA:LI:DA:DD:RE is not
    """
    try:
        pairs = address.split(":")
        if len(pairs) != 6: return False
        for b in pairs: int(b, 16)
    except:
        return False
    return True

def get_available_port( protocol ):
    """
    finds and returns an available port for the specified protocol.  protocol 
    must be either L2CAP or RFCOMM, and None if no available port is found.

    It is possible that somebody will bind to the returned port in between the
    time it is returned and the time you use it, so be careful.

    This is a major hack.  In Linux kernel versions 2.6.7 and higher, you can
    use PSM/channel 0 when binding a socket to be dynamically assigned a port
    number (e.g.  sock.bind(("",0)) )
    """
    if protocol == RFCOMM:
        for channel in range(1,31):
            s = BluetoothSocket( RFCOMM )
            try:
                s.bind( ("", channel) )
                s.close()
                return channel
            except:
                s.close()
    elif protocol == L2CAP:
        for psm in range(0x1001,0x8000,2):
            s = BluetoothSocket( L2CAP )
            try:
                s.bind( ("", psm) )
                s.close()
                return psm
            except:
                s.close()
    else:
        raise ValueError("protocol must either RFCOMM or L2CAP")

_socketmethods = ( 'bind', 'connect', 'connect_ex', 'close',
        'dup', 'fileno', 'getpeername', 'getsockname', 'gettimeout',
        'getsockopt', 'listen', 'makefile', 'recv', 'recvfrom', 'sendall',
        'send', 'sendto', 'setblocking', 'setsockopt', 'settimeout', 'shutdown'
        )

class BluetoothSocket:
    __doc__ = _bluetooth.btsocket.__doc__

    def __init__(self, proto = RFCOMM, _sock=None):
        if _sock is None:
            _sock = bt.btsocket(proto)
        self._sock = _sock
        self._proto = proto

    def dup(self):
        """dup() -> socket object

        Return a new socket object connected to the same system resource."""
        return BluetoothSocket(proto=self._proto, _sock=self._sock)

    def accept(self):
        try:
            client, addr = self._sock.accept()
        except _bluetooth.error, e:
            raise BluetoothError(str(e))
        newsock = BluetoothSocket(self._proto, client)
        return (newsock, addr)
    accept.__doc__ = _bluetooth.btsocket.accept.__doc__

    # import methods from the wraapped socket object
    _s = ("""def %s(self, *args, **kwargs): 
    try: 
        return self._sock.%s(*args, **kwargs)
    except _bluetooth.error, e:
        raise BluetoothError(str(e))
%s.__doc__ = _bluetooth.btsocket.%s.__doc__\n""")
    for _m in _socketmethods:
        exec _s % (_m, _m, _m, _m)
    del _m, _s
    
class DeviceDiscoverer:
    """
    Skeleton class for finer control of the device discovery process.

    To implement asynchronous device discovery (e.g. if you want to do 
    something *as soon as* a device is discovered), subclass DeviceDiscoverer
    and override device_discovered() and inquiry_complete()
    """
    def __init__(self):
        self.sock = None
        self.is_inquiring = False
        self.lookup_names = False

        self.names_to_find = {}
        self.names_found = {}

    def find_devices(self, lookup_names=True, 
            duration=8, 
            flush_cache=True):
        """
        start_inquiry( lookup_names=True, service_name=None, 
                       duration=8, flush_cache=True )

        Call this method to initiate the device discovery process

        lookup_names - set to True if you want to lookup the user-friendly 
                       names for each device found.

        service_name - set to the name of a service you're looking for.
                       only devices with a service of this name will be 
                       returned in device_discovered()


        ADVANCED PARAMETERS:  (don't change these unless you know what 
                            you're doing)

        duration - the number of 1.2 second units to spend searching for
                   bluetooth devices.  If lookup_names is True, then the 
                   inquiry process can take a lot longer.

        flush_cache - return devices discovered in previous inquiries
        """
        if self.is_inquiring:
            raise BluetoothError("Already inquiring!")

        self.lookup_names = lookup_names

        self.sock = _gethcisock()
        flt = bt.hci_filter_new()
        bt.hci_filter_all_events(flt)
        bt.hci_filter_set_ptype(flt, bt.HCI_EVENT_PKT)

        try:
            self.sock.setsockopt( bt.SOL_HCI, bt.HCI_FILTER, flt )
        except:
            raise BluetoothError("problem with local bluetooth device.")

        # send the inquiry command
        duration = 4
        max_responses = 255
        cmd_pkt = struct.pack("BBBBB", 0x33, 0x8b, 0x9e, \
                duration, max_responses)

        self.pre_inquiry()
        
        try:
            bt.hci_send_cmd( self.sock, bt.OGF_LINK_CTL, \
                    bt.OCF_INQUIRY, cmd_pkt)
        except:
            raise BluetoothError("problem with local bluetooth device.")

        self.is_inquiring = True

        self.names_to_find = {}
        self.names_found = {}

    def cancel_inquiry(self):
        """
        Call this method to cancel an inquiry in process.  inquiry_complete 
        will still be called.
        """
        try:
            bt.hci_send_cmd( self.sock, bt.OGF_LINK_CTL, \
                    bt.OCF_INQUIRY_CANCEL )
            self.names_to_find = {}
        except:
            raise BluetoothError("error canceling inquiry")

    def process_inquiry(self):
        """
        Repeatedly calls process_event() until the device inquiry has 
        completed.
        """
        while self.is_inquiring or len(self.names_to_find) > 0:
            self.process_event()
        print "done"

    def process_event(self):
        """
        Waits for one event to happen, and proceses it.  The event will be
        either a device discovery, or an inquiry completion.
        """
        self._process_hci_event()

    def _process_hci_event(self):
        # voodoo magic!!!
        pkt = self.sock.recv(255)
        ptype, event, plen = struct.unpack("BBB", pkt[:3])
        pkt = pkt[3:]
        if event == bt.EVT_INQUIRY_RESULT:
            nrsp = struct.unpack("B", pkt[0])[0]
            for i in range(nrsp):
                addr = bt.ba2str( pkt[1+6*i:1+6*i+6] )
                psrm = pkt[ 1+6*nrsp+i ]
                pspm = pkt[ 1+7*nrsp+i ]
                devclass_raw = pkt[1+9*nrsp+3*i:1+9*nrsp+3*i+3]
                devclass = struct.unpack("!I", "\0%s" % devclass_raw)[0]
                clockoff = pkt[1+12*nrsp+2*i:1+12*nrsp+2*i+2]

                self._device_discovered( addr, devclass, psrm, pspm, clockoff )
        elif event == bt.EVT_INQUIRY_RESULT_WITH_RSSI:
            nrsp = struct.unpack("B", pkt[0])[0]
            for i in range(nrsp):
                addr = bt.ba2str( pkt[1+6*i:1+6*i+6] )
                psrm = pkt[ 1+6*nrsp+i ]
                pspm = pkt[ 1+7*nrsp+i ]
                devclass_raw = pkt[1+8*nrsp+3*i:1+8*nrsp+3*i+3]
                devclass = struct.unpack("!I", "\0%s" % devclass_raw)[0]
                clockoff = pkt[1+11*nrsp+2*i:1+11*nrsp+2*i+2]
                rssi = struct.unpack("b", pkt[1+13*nrsp+i])[0]

                self._device_discovered( addr, devclass, psrm, pspm, clockoff )
        elif event == bt.EVT_INQUIRY_COMPLETE:
            self.is_inquiring = False
            if len(self.names_to_find) == 0:
#                print "inquiry complete (evt_inquiry_complete)"
                self.sock.close()
                self.inquiry_complete()
            else:
                self._send_next_name_req()

        elif event == bt.EVT_CMD_STATUS:
            status, ncmd, opcode = struct.unpack("BBH", pkt[:4])
            if status != 0:
                self.is_inquiring = False
                self.sock.close()
                
#                print "inquiry complete (bad status 0x%X 0x%X 0x%X)" % \
#                        (status, ncmd, opcode)
                self.names_to_find = {}
                self.inquiry_complete()
        elif event == bt.EVT_REMOTE_NAME_REQ_COMPLETE:
            status = struct.unpack("B", pkt[0])[0]
            addr = bt.ba2str( pkt[1:7] )
            if status == 0:
                name = pkt[7:]
                if addr in self.names_to_find:
                    device_class = self.names_to_find[addr][0]
                    self.device_discovered( addr, device_class, name )
                    del self.names_to_find[addr]
                    self.names_found[addr] = ( device_class, name )
                else:
                    pass
            else:
                if addr in self.names_to_find: del self.names_to_find[addr]
                # XXX should we do something when a name lookup fails?
#                print "name req unsuccessful %s - %s" % (addr, status)

            if len(self.names_to_find) == 0:
                self.is_inquiring = False
                self.sock.close()
                self.inquiry_complete()
#                print "inquiry complete (name req complete)"
            else:
                self._send_next_name_req()
        else:
            pass
#            print "unrecognized packet type 0x%02x" % ptype

    def _device_discovered(self, address, device_class, psrm, pspm, clockoff):
        if self.lookup_names:
            if address not in self.names_found and \
                address not in self.names_to_find:
            
                self.names_to_find[address] = \
                    (device_class, psrm, pspm, clockoff)
        else:
            self.device_discovered(address, device_class, None)

    def _send_next_name_req(self):
        assert len(self.names_to_find) > 0
        address = self.names_to_find.keys()[0]
        device_class, psrm, pspm, clockoff = self.names_to_find[address]
        bdaddr = bt.str2ba( address )
        
        cmd_pkt = "%s%s\0%s" % (bdaddr, psrm, clockoff)

        try:
            bt.hci_send_cmd( self.sock, bt.OGF_LINK_CTL, \
                    bt.OCF_REMOTE_NAME_REQ, cmd_pkt)
        except Exception, e:
            raise BluetoothError("error request name of %s - %s" % 
                    (address, str(e)))

    def fileno(self):
        if not self.sock: return None
        return self.sock.fileno()

    def pre_inquiry(self):
        """
        Called just after start_inquiry is invoked, but just before the inquiry
        is started.

        This method exists to be overriden
        """

    def device_discovered(self, address, device_class, name):
        """
        Called when a bluetooth device is discovered.

        address is the bluetooth address of the device

        device_class is the Class of Device, as specified in [1]
                     passed in as a 3-byte string

        name is the user-friendly name of the device if lookup_names was set
             when the inquiry was started.  otherwise None
        
        This method exists to be overriden.

        [1] https://www.bluetooth.org/foundry/assignnumb/document/baseband
        """
        if name:
            print "found: %s - %s (class 0x%X)" % (address, name, device_class)
        else:
            print "found: %s (class 0x%X)" % (address, device_class)

    def inquiry_complete(self):
        """
        Called when an inquiry started by start_inquiry has completed.
        """
        print "inquiry complete"

# ============== SDP service registration and unregistration ============
def advertise_service( sock, name, uuid = "" ):
    """
    Advertises a service with the local SDP server.  sock must be a bound,
    listening socket.  name should be the name of the service, and uuid (if
    specified) should be a string of the form "XXXX-XXXX-XXXX-XXXX", where
    each 'X' is a hexadecimal digit.
    """
    try:
        _bluetooth.sdp_advertise_service( sock._sock, name, uuid )
    except _bluetooth.error, e:
        raise BluetoothError(str(e))

def stop_advertising( sock ):
    """
    Instructs the local SDP server to stop advertising the service associated
    with sock.  You should typically call this right before you close sock.
    """
    try:
        _bluetooth.sdp_stop_advertising( sock._sock )
    except _bluetooth.error, e:
        raise BluetoothError(str(e))

def find_service( name = None, uuid = None, address = None ):
    """
    find_service( name = None, uuid = None, address = None )

    Searches for SDP services that match the specified criteria and returns
    the search results.  If no criteria are specified, then returns a list of
    all nearby services detected.  If more than one is specified, then
    the search results will match all the criteria specified.  If uuid is
    specified, it must be a string of the form "XXXX-XXXX-XXXX-XXXX", where
    each 'X' is a hexadecimal digit.  A special case of address is
    "localhost", which will search for services on the local machine.

    The search results will be a list of dictionaries.  Each dictionary
    represents a search match and will have the following key/value pairs:

      host - the bluetooth address of the device advertising the service
      name - the name of the service being advertised
      description - a description of the service being advertised
      provider - the name of the person/organization providing the service
      protocol - either 'RFCOMM', 'L2CAP', None if the protocol was not
                 specified, or 'UNKNOWN' if the protocol was specified but 
                 unrecognized
      port - the L2CAP PSM # if the protocol is 'L2CAP', the RFCOMM channel #
             if the protocol is 'RFCOMM', or None if it wasn't specified
    """
    if not address:
        devices = discover_devices()
    else:
        devices = [ address ]

    results = []

    try:
        for addr in devices:
            s = _bluetooth.SDPSession()
            s.connect( addr )
            matches = []
            if uuid is not None:
                matches = s.search(uuid)
            else:
                matches = s.browse()

            if name is not None:
                matches = filter( lambda s: s.get( "name", "" ) == name, \
                        matches )

            for m in matches:
                m["host"] = addr

            results.extend(matches)
    except _bluetooth.error, e:
        raise BluetoothError(str(e))

    return results
    
# ================= internal methods =================
def _gethcisock( device_id = -1):
    try:
        sock = bt.hci_open_dev( device_id )
    except:
        raise BluetoothError("error accessing bluetooth device")
    return sock

def _checkaddr(addr):
    if not is_valid_address(addr): 
        raise BluetoothError("%s is not a valid Bluetooth address" % addr)


if __name__ == "__main__":
    print "performing device discovery"
    print discover_devices()
